
/* Generated by modes.py; do not edit! */

/*
 * Copyright © 2018, 2020 Christian Persch
 *
 * This library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library.  If not, see <https://www.gnu.org/licenses/>.
 */

#if !defined(MODE) || !defined(MODE_FIXED)
#error "Must define MODE and MODE_FIXED before including this file"
#endif

MODE_FIXED(GATM, 1, ALWAYS_RESET)
MODE_FIXED(KAM, 2, ALWAYS_RESET)
MODE_FIXED(CRM, 3, ALWAYS_RESET)
MODE(IRM, 4)
MODE_FIXED(SRTM, 5, ALWAYS_RESET)
MODE_FIXED(ERM, 6, ALWAYS_RESET)
MODE_FIXED(VEM, 7, ALWAYS_RESET)
MODE(BDSM, 8)
MODE_FIXED(DCSM, 9, ALWAYS_SET)
MODE_FIXED(HEM, 10, ALWAYS_RESET)
MODE_FIXED(PUM, 11, ALWAYS_RESET)
MODE_FIXED(SRM, 12, ALWAYS_SET)
MODE_FIXED(FEAM, 13, ALWAYS_RESET)
MODE_FIXED(FETM, 14, ALWAYS_RESET)
MODE_FIXED(MATM, 15, ALWAYS_RESET)
MODE_FIXED(TTM, 16, ALWAYS_RESET)
MODE_FIXED(SATM, 17, ALWAYS_RESET)
MODE_FIXED(TSM, 18, ALWAYS_RESET)
MODE_FIXED(EBM, 19, ALWAYS_RESET)
MODE_FIXED(LNM, 20, ALWAYS_RESET)
MODE_FIXED(GRCM, 21, ALWAYS_SET)
MODE_FIXED(ZDM, 22, ALWAYS_RESET)
MODE_FIXED(WYDSCM, 30, ALWAYS_RESET)
MODE_FIXED(WYSTLINM, 31, ALWAYS_RESET)
MODE_FIXED(WYCRTSAVM, 32, ALWAYS_RESET)
MODE_FIXED(WYSTCURM, 33, ALWAYS_RESET)
MODE_FIXED(WYULCURM, 34, ALWAYS_RESET)
MODE_FIXED(WYCLRM, 35, ALWAYS_SET)
MODE_FIXED(WYDELKM, 36, ALWAYS_RESET)
MODE_FIXED(WYGATM, 37, ALWAYS_RESET)
MODE_FIXED(WYTEXM, 38, ALWAYS_RESET)
MODE_FIXED(WYEXTDM, 40, ALWAYS_SET)
MODE_FIXED(WYASCII, 42, ALWAYS_SET)
